case_study = """path where the case study directory where custom classes for
organism, environment, experiment, observation have been specified in a mod.py
module and a simulation has been specified in a sim.py module which creates and
runs a simulation"""

scenario = """a string of the scenario specified for the case studies, which
contains folders with the config files for the scenario like events.csv,
config.json and sbi.json"""

package = """path relative to working directory ($PWD) where case studies are 
located"""


import socket
import os
import pytest


class TestSbiPipeline:
    """tests sbi pipeline"""
    case_study = "core_daphnia"
    scenario = "test"

    def test_generate_sims(self):
        pytest.skip()   
        from pymob import generate_sims
        print(f"hostname: {socket.gethostname()}, user: {os.environ.get('USERNAME')}")
        generate_sims.main(
            case_study=[self.case_study, self.scenario],
            n_sims=5,
            worker="1"
        )
        generate_sims.main(
            case_study=[self.case_study, self.scenario],
            n_sims=5,
            worker="2"
        )

    def test_process_simulations(self):
        pytest.skip()   
        from pymob import process_simulations
        print(f"hostname: {socket.gethostname()}, user: {os.environ.get('USERNAME')}")
        process_simulations.main(
            case_study=[self.case_study, self.scenario],
        )

    def test_prior_predictive_checks(self):
        pytest.skip()   
        from pymob import prior_predictive_checks
        prior_predictive_checks.main(
            case_study=[self.case_study, self.scenario]
        )

    def test_train_network_SNPE(self):
        pytest.skip()   
        from pymob import train_network
        print(f"hostname: {socket.gethostname()}, user: {os.environ.get('USERNAME')}")
        train_network.main(
            case_study=[self.case_study, self.scenario],
            inferer_engine="SNPE", 
            training_batch_size=1
        )

    def test_train_network_SNLE(self):
        pytest.skip()   
        from pymob import train_network
        print(f"hostname: {socket.gethostname()}, user: {os.environ.get('USERNAME')}")
        train_network.main(
            case_study=[self.case_study, self.scenario],
            inferer_engine="SNLE", 
            training_batch_size=1
        )

    def test_snle_sampling(self):
        pytest.skip()   
        from pymob.inference.sbi import sbi_snle_sample_posterior
        print(f"hostname: {socket.gethostname()}, user: {os.environ.get('USERNAME')}")
        sbi_snle_sample_posterior.main(
            case_study=[self.case_study, self.scenario],
        )


    def test_evaluate_sbi(self):
        pytest.skip()   
        from pymob import evaluate_sbi
        evaluate_sbi.main(
            case_study=[self.case_study, self.scenario],
        )

    def test_posterior_predictions(self):
        pytest.skip()   
        from pymob import posterior_predictions
        posterior_predictions.main(
            case_study=[self.case_study, self.scenario],
            posterior_cluster=0,
            nsims=2
        )

    def test_plot_posterior_predictions(self):
        pytest.skip()   
        from pymob import plot_posterior_predictions
        plot_posterior_predictions.main(
            case_study=[self.case_study, self.scenario],
            posterior_cluster=0
        )


from scipy.optimize import basinhopping
from sklearn.preprocessing import PowerTransformer, QuantileTransformer, MinMaxScaler
from data.datasets import indy
from pymob.utils.store_file import read_config
from sims.simulation import create_sim, Simulation
from matplotlib import pyplot as plt
import os
import json
import pandas as pd
import numpy as np
import sys
import shutil        


class IndyOptimizer():
    def __init__(
        self, 
        f,
        params,
        niter,
        temperature,
        stepsize,
        interval,
        minimizer_kwargs={},
        sim_config="config/parameters/optimization_indy.json",
        optimize_on_features="all",
        optimize_on_sample=0,
        seed=None
        ):
        self.f = f
        self.parnames = []
        self.x0 = []
        self.bounds = []
        self.niter = niter
        self.temperature = temperature
        self.stepsize = stepsize
        self.interval = interval
        self.steps = []
        self.seed = seed
        self.minimizer_kwargs = minimizer_kwargs
        self.optimize_on_features=optimize_on_features
        self.optimize_on_sample=optimize_on_sample
        self.parse_params(params)
        
        self.data = indy()
        self.hist = {"x":[], "f": [], "a":[]}
        self.result = None
        self.x = None
        self.sim_config = read_config(sim_config)
        self.test_sim = create_sim(self.sim_config)
        self.test_sim.run()

        self.use_sim_cols = self.find_sim_cols()
        self.use_exp_cols = self.find_exp_cols()

        self.scaler = self.construct_scaler()
        self.parameter_scales = self.harmonize_scales()
        self.prepare_data()
        self.copy_config()
        print("--- initialized optimization ---")
        print("optimizing for:", optimize_on_features)
        print("optimizing parameters:", self.parnames)


    class Trap:
        """
        Escape Minimization if the same local minimum is encountered twice in a 
        row (happens due to non-continuous energy landscapes)
        """

        def __init__(self):
            pass
            # Avoid ZeroDivisionError since "MBH can be regarded as a special case
            # of the BH framework with the Metropolis criterion, where temperature
            # T = 0." (Reject all steps that increase energy.)

        def accept_reject(self, energy_new, energy_old):
            """
            if new energy is equal to old energy, reject the test, because the
            minimum was reached before and most of the time means that 
            the optimizer is trapped in a far off region in a low energy landscape
            accept (return true) if energy new is unequal to energy old
            """
            return energy_new != energy_old

        def __call__(self, **kwargs):
            """
            f_new and f_old are mandatory in kwargs
            """
            return bool(self.accept_reject(kwargs["f_new"],
                        kwargs["f_old"]))


    class TakeCustomSteps:
        def __init__(self, stepsize=1, steps=[]):
            self.stepsize = stepsize
            self.rng = np.random.default_rng()
            self.steps = steps

        def prepare_steps(self, x):
            if len(self.steps) == 0:
                return np.array(np.repeat(self.stepsize, len(x)))

            return self.stepsize * np.array(self.steps)

        def __call__(self, x):
            s = self.prepare_steps(x)
            x += self.rng.uniform(-s, +s)
            return x

    def parse_params(self, params):
        # parse initial guess of x
        for key, value in params.items():
            self.parnames.append(key)
            self.x0.append(value[0])

        self.x0 = np.array(self.x0)
        scales = self.harmonize_scales()

        # parse bounds and steps
        for (key, value), scal in zip(params.items(), scales):
            lb = value[1][0] / scal
            ub = value[1][1] / scal
            try:
                s = value[2] / scal
                self.steps.append(s)
            except IndexError:
                # if no steps are used take step to the initial value times stepsize
                s = value[0] / scal * 0.5
                self.steps.append(s)

            self.bounds.append((lb, ub))
        
    def find_exp_cols(self):
        labels = self.data["labels"]["values"]
        if self.optimize_on_features == "all":
            self.optimize_on_features = labels
            return [True] * len(labels)
        return [True if l in self.optimize_on_features else False for l in labels]

    def find_sim_cols(self):
        labels = self.test_sim.events.observations["labels"][1:]
        if self.optimize_on_features == "all":
            self.optimize_on_features = labels
            return [True] * len(labels)
        return [True if l in self.optimize_on_features else False for l in labels]

    def prepare_data(self):
        self.exp_sample = self.data["data"][:, self.optimize_on_sample, self.use_exp_cols]
        # mask = np.all(np.isnan(exp_sample), axis=1)
        # self.exp_sample = exp_sample[~mask]
        self.exp_sample_transformed = self.scaler.transform(self.exp_sample)
        
    def construct_scaler(self):
        # TODO: scaling is still a bit problematic, because they affect the fitting procedure
        #       and different transformers result in different weights
        #       sometimes large offspring data has almost no higher effect over small offspring (yeo-johnson)
        #       data under quantile transformer it is a bit disproportionally high
        #       In addition: scaling under quantile transformer optimizes heavily for 
        #       offspring frequency and not for quantitiy (has almost no effect)
        
        # scaler = QuantileTransformer(random_state=0, output_distribution="normal")
        # scaler = PowerTransformer(method="yeo-johnson")
        # new_shape = (self.data["shape"]["day"] * self.data["shape"]["sample"], sum(self.use_exp_cols))
        # data = self.data["data"][:, :, self.use_exp_cols].reshape(new_shape)
        new_shape = (self.data["shape"]["day"] , sum(self.use_exp_cols))
        data = self.data["data"][:, self.optimize_on_sample, self.use_exp_cols].reshape(new_shape)
        data = np.row_stack((data, np.zeros(sum(self.use_exp_cols))))

        scaler = MinMaxScaler().fit(data)
        return scaler

    def harmonize_scales(self):
        return 10 ** np.ceil(np.log10(self.x0))

    def callback(self, x, f, accepted):
        self.hist["x"].append(x * self.parameter_scales)
        self.hist["f"].append(f)
        self.hist["a"].append(int(accepted))
        # update trajectory as it goes on (helpful if bh optimizer is terminated prematurely)
        self.save_trajectory() 
        print("at minimum %.4f accepted %d" % (f, int(accepted)), x * self.parameter_scales, file=sys.stdout)

    def optim(self):
        config = self.sim_config
        scales = self.parameter_scales
        use_sim_cols = self.use_sim_cols
        scaler = self.scaler
        exp_data = self.exp_sample_transformed
        parnames = self.parnames
        minimizer_kwargs = self.minimizer_kwargs.copy()

        minimizer_kwargs.update({
            "bounds": self.bounds,
            "args": (config, scales, use_sim_cols, exp_data, scaler, parnames)   
        })

        f = self.f
        x0 = self.x0 / scales
        temperature = self.temperature
        stepsize = self.stepsize
        interval = self.interval
        accept_test = self.Trap()
        take_step = self.TakeCustomSteps(stepsize, self.steps)
        niter = self.niter
        callback = self.callback
        seed = self.seed

        self.result = basinhopping(
            f, x0, niter, 
            T=temperature,
            # stepsize=stepsize, # only needed if no custom step taking algorithm was defined
            take_step=take_step,
            interval=interval,
            minimizer_kwargs=minimizer_kwargs,
            callback=callback,  
            accept_test=accept_test,
            disp=True,
            seed=seed,
        )

        self.x = self.result.x * self.parameter_scales

    def copy_config(self):
        path = self.test_sim.directory
        self.test_sim.create_directory()

        # copy config
        sim_conf = self.sim_config
        with open(os.path.join(path, "sim_conf.json"), "w") as f:
            json.dump(sim_conf, f, indent=4)

        # copy eventfile
        eventfile = sim_conf["events"]["eventfile"]
        shutil.copy(eventfile, os.path.join(path, os.path.basename(eventfile)))

    def store_data(self):
        instance = self.__dict__
        data = {}
        keys = [
            "x0", "bounds", "niter", "seed", "temperature", "stepsize",
            "minimizer_kwargs",
            "optimize_on_features", "optimize_on_sample", 
            "sim_config", "parameter_scales", "x"
        ]

        for k in keys:
            values = instance[k]
            if isinstance(values, np.ndarray):
                values = list(values)
            
            data.update({k: values})

        self.test_sim.create_directory()
        with open(os.path.join(self.test_sim.directory, "optim_params.json"), "w") as f:
            json.dump(data, f, indent=4)

        self.save_trajectory()

    def save_trajectory(self):
        self.test_sim.create_directory()
        X = pd.DataFrame.from_records(self.hist["x"], columns=self.parnames)
        F = pd.DataFrame(self.hist["f"], columns=["sse"])
        A = pd.DataFrame(self.hist["a"], columns=["accept"])

        D = pd.concat([X,F,A],axis=1)

        D.to_csv(os.path.join(self.test_sim.directory, "optim_trajectory.csv"), index=False)

    def plot_results(self, sim):
        """
        sim:    simulation instance (which has already run or configuration for sim)
        """
        if not isinstance(sim, Simulation):
            s = create_sim(sim)
            s.run()
        sim_data = s.events.get_tensor()[:, 0, self.use_sim_cols]
        exp_data = self.exp_sample
        feature_names = self.optimize_on_features
        scaler = self.scaler
        xlabels = self.data["labels"]["day"]
        sse = self.result.fun

        plot_fit(exp_data, sim_data, feature_names, scaler, xlabels, sse)
        self.test_sim.create_directory()
        plt.savefig(os.path.join(self.test_sim.directory, "optim_result.png"))

def plot_fit(exp_data, sim_data, feature_names, scaler, xlabels, sse):
    exp_sample_transformed = scaler.transform(exp_data)
    sim_sample_transformed = scaler.transform(
        np.nan_to_num(sim_data, nan=0))
    diff = exp_sample_transformed - sim_sample_transformed
    n_features = len(feature_names)
    sse_contrib = np.sum(np.nan_to_num(diff**2, nan=0), axis=0)
    fig, axes = plt.subplots(ncols=n_features, nrows=1, figsize=(n_features*4, 5))

    for i, ax, label in zip(range(sim_data.shape[1]), axes, feature_names):
        ax.scatter(xlabels, np.nan_to_num(
            sim_data[:, i]), label="simulation", alpha=.5)
        ax.scatter(xlabels, exp_data[:, i], label="experiment", alpha=.5)

        if label == "survival":
            ax.set_ylim(-0.05, 1.05)
        ax.set_ylabel(label)
        ax.set_xlabel("time [days]")

        ax.vlines(x=xlabels,
                    ymin=np.nan_to_num(sim_data[:, i]),
                    ymax=exp_data[:, i],
                    linestyle="--", color="black")

        # add sse contributions
        ax.text(0.99, 0.99,
            s="sse: {}".format(round(sse_contrib[i], 2)),
                transform=ax.transAxes, va="top", ha="right")

        if i == 0:
            ax.text(0.0, 1.01, s="overall error: {}".format(sse),
                    transform=ax.transAxes)

    axes[0].legend()




# visit http://127.0.0.1:8050/ in your web browser.
import json
import glob
import os
import numpy as np
import xarray as xr
from dash import Dash, html, dcc, Input, Output, ALL
from plotly.subplots import make_subplots
from plotly.colors import sample_colorscale, qualitative
import plotly.graph_objects as go
from dash_bootstrap_components import (
    Row, Col, Container, themes, Card, CardBody, Tab, Tabs)
from pymob.utils import store_file as store
from pymob.sims.simulation import update_parameters

CASE_STUDIES = "case_studies"
assert os.path.exists(CASE_STUDIES), "make sure you have a 'case_studies' directory in your project directory"


app = Dash(__name__, external_stylesheets=[themes.BOOTSTRAP])

models = dict()



case_studies = [os.path.split(cs)[-1] for cs in glob.glob(f"{CASE_STUDIES}/*")]

persist_kw = dict(persistence=True, persistence_type="session")
substance_drop = ["diuron", "naproxen", "diclofenac"]

# headline and some intro text
intro = Row(Col([
    html.H1(children='Interactive Simulation'),
    html.Div(children="""
        An interactive user interface for testing timepath
        simulation before going large scale simulations"""),
    html.Br(),
    html.H2(children="select a case study and a scenario"),
    html.Br(),
]))

# case study and scenario selection
case_study_selection = Row([
    Col(html.Div([
        dcc.Dropdown(
            options=case_studies, value=case_studies[0],
            id='case-study', **persist_kw),
        dcc.Dropdown(id='scenario', value="test", **persist_kw),
    ])),
    # substance selection
    Col(html.Div([
        dcc.Dropdown(options=substance_drop, id='substance-dropdown', 
            value=substance_drop[0], **persist_kw),
    ]))
])

# storage
storage = Row(Col([
    html.Div([
        dcc.Store(id='config-file'),
        dcc.Store(id='config-flat'),
        dcc.Store(id='parameter-keys'),
        dcc.Store(id='data-observed'),
        dcc.Store(id='data-simulated'),
    ])
]))
# parameter sliders
parameters = html.Div(id='parameter-input')


concentration_plot = Tab(Card(
    CardBody(
        html.Div([
            dcc.Graph(
                id='fig-concentration', 
                mathjax=True,
                style={'height': '90vh'}
            ),
        ]),
    )
), label="concentration")

simulation = Row(Col([
    html.Br(),
    Row([
        Col([
            html.H3("Set Parameters"),
            parameters
        ], width=4),
        Col([
            html.H3("Results"),
            dcc.Dropdown(id="observation-ids", multi=True, searchable=True, **persist_kw) ,
            Tabs(concentration_plot)
        ], width=8),
    ]),
]))

app.layout = Container([
    intro, 
    case_study_selection,
    simulation,
    storage,
], fluid=True)


@app.callback(
    Output("scenario", "options"),
    Input("case-study", "value")
)
def set_scenario_options(case_study):
    search = os.path.join(CASE_STUDIES, case_study, "scenarios", "*")
    scenarios = [os.path.split(s)[-1] for s in glob.glob(search)]
    return [{'label': i, 'value': i} for i in scenarios]


# handling scenario and case study loading
@app.callback(
    Output("config-file", "data"),
    Input("case-study", "value"),
    Input("scenario", "value"),
)
def load_study(case_study, scenario):
    # since this is computationally intensive, only load models once
    config = store.prepare_scenario(case_study, scenario)
    mod = store.import_package(package_path=config["simulation"]["package"])
    models[case_study] = mod

    return json.dumps(config)


@app.callback(
    Output('parameter-input', 'children'), 
    Output('config-flat', 'data'), 
    Output('parameter-keys', 'data'), 
    Input('config-file', 'data'),
    # Input('substance-dropdown', 'value')
)
def set_parameters(cfg_serialized):
    # config = store.prepare_scenario(case_study="tktd_mix", scenario="diuron_metabo_pyabc")
    config = json.loads(cfg_serialized)

    # how to clear this list again: see 
    # https://dash.plotly.com/pattern-matching-callbacks

    parameter_keys = []
    sections = []
    for key, config_section in config.items():
        section_params = list()
        flat_config = store.unnest(config_section, [])
        for param, value in flat_config:
            par_id = f"{key}.{param}"
            input_id = {"type": "parameter-value", "index": par_id}
            if isinstance(value, bool):
                inp = dcc.Input(
                    value=int(value),
                    type="number",
                    id=input_id,
                    **persist_kw
                )
            elif isinstance(value, (float, int)):
                inp = dcc.Input(
                    value=value, 
                    type="number",
                    id=input_id, 
                    **persist_kw
                )
                # sliders have the problem that they cannot be used displayed if 
                # values are too low. This behaviour is shit.
                # Im starting to favour normal inputs, because they are also 
                # independent of scale
                # or use log inputs and let sliders be on the log scale with 
                # appropriate marks 10 ** -1, .. 10 ** 2
                # and then transform all sliders back to 10 ** x
                # transformation of float to parameters that need to be int 
                # should be done inside the sim
            elif isinstance(value, str):
                inp = dcc.Input(
                    value=value,
                    type="text",
                    id=input_id,
                    **persist_kw
                )
            elif isinstance(value, (list, tuple)):
                continue

            else:
                raise NotImplementedError(
                    f"the config type {type(value)} of {par_id} cannot be parsed.")

            parcol = Row(Col([
                html.Div(param),
                inp
            ]))
            section_params.append(parcol)
            parameter_keys.append(par_id)

        tab = Tab(
            Card(CardBody(
                section_params)
        ), label=key)

        sections.append(tab)

    return (
        Tabs(children=sections), 
        json.dumps(flat_config), 
        json.dumps(parameter_keys)
    )


# load observed dataset
@app.callback(
    Output("data-observed", "data"),
    Output("observation-ids", "options"),
    Input("case-study", "value"),
    Input("config-file", "data"), # is only there to delay the callback of data
    Input("substance-dropdown", "value"),
)
def load_data(case_study, config_file, substance):
    data = models[case_study].data

    # this method needs to be specified via configuration
    obs_data = data.substance_tk_dataset(substance)
    obs_ids = list(obs_data.id.values)
    datas = store.serialize(obs_data, convert_time=True)
    return datas, [{'label': i, 'value': i} for i in obs_ids]


# simulate dataset
@app.callback(
    Output('data-simulated', 'data'),
    Input("case-study", "value"),
    Input("config-file", "data"),
    Input("parameter-keys", "data"),
    Input({"type": "parameter-value", "index": ALL}, "value"),
    Input("observation-ids", "value"),
    Input("data-observed", "data"),
)
def simulate(case_study, cfg_serialized, param_keys, param_values, obs_ids, obs_datas):
    config = update_parameters(
        json.loads(cfg_serialized), param_values, json.loads(param_keys))
    obs_data = store.deserialize(obs_datas)

    Simulation = models[case_study].sim.simulation
    config_modifier = models[case_study].sim.data_to_config
    
    # init and run sim, extract observations
    simulation_output = []
    for i in obs_ids:
        config = config_modifier(dataset=obs_data.sel(id=i), cfg=config, save_event_file=True)
        sim = Simulation(config)
        sim.run()
        sim_out = sim.experiment.get_tensor(return_time=True, return_xarray=True)
        sim_out = sim_out.assign_coords(id=i)

        simulation_output.append(sim_out)

    sim_data = xr.concat(simulation_output, dim="id")

    return store.serialize(sim_data, convert_time=True)


# finally plot figure
@app.callback(
    Output("fig-concentration", "figure"),
    Input("data-simulated", "data"),
    Input("data-observed", "data"),
    Input("observation-ids", "value"),
)
def fig_concentration(sim_datas, obs_datas, observation_ids):
    sim_data = store.deserialize(sim_datas)
    obs_data = store.deserialize(obs_datas)
    
    fig = make_subplots(rows=2, cols=1, shared_xaxes=True, vertical_spacing=0.1)
    
    ns2h = 1e-9 / 3600
    colors = qualitative.D3
    for i, c in zip(observation_ids, colors):
        # plot simulated data
        fig.add_trace(
            go.Scatter(
                x=sim_data.time * ns2h, 
                y=sim_data.cext_diuron.sel(id=i), 
                name=f"sim {i}",
                mode="lines", line_color=c),
            row=1, col=1)
        fig.add_trace(
            go.Scatter(
                x=sim_data.time * ns2h, 
                y=sim_data.cint_diuron.sel(id=i), 
                name=f"sim {i}",
                showlegend=False,
                mode="lines", line_color=c),
            row=2, col=1)

        # plot observed data
        fig.add_trace(
            go.Scatter(
                x=obs_data.time * ns2h, 
                y=obs_data.cext.sel(id=i), 
                name=f"obs {i}",
                mode="markers", marker_color=c),
            row=1, col=1)
        fig.add_trace(
            go.Scatter(
                x=obs_data.time * ns2h, 
                y=obs_data.cint.sel(id=i), 
                showlegend=False,
                name=f"obs {i}",
                mode="markers", marker_color=c),
            row=2, col=1)
        # fig.update_traces(mode="markers")

    fig.update_layout(transition_duration=500, template="simple_white")

    x_kwargs = dict(
        showgrid=False,
        mirror=True,
        # range=[float(sim_data.time.min()), float(sim_data.time.max()) * ns2h]
    )
    
    y_kwargs = dict(
        showgrid=False,
        mirror=True,
    )

    fig.update_xaxes(title="Time [h]", row=2, col=1, **x_kwargs)
    fig.update_xaxes(title=None, row=1, col=1, **x_kwargs)
    fig.update_yaxes(title=r"$C_e$", row=1, col=1, **y_kwargs)
    fig.update_yaxes(title=r"$C_i$", row=2, col=1, **y_kwargs)
    return fig


    # debugging issues https://stackoverflow.com/questions/22711087/flask-importerror-no-module-named-app



if __name__ == "__main__":
    app.run_server(debug=False)
# from __future__ import print_function
# from ipywidgets import interact, interactive, fixed, interact_manual
# import ipywidgets as widgets


# from objects.environments import World
# from objects.organisms import Daphnia
# from sims.simulation import create_sim
# from helpers.store_file import read_config


# # read parameter file
# def run_sim(
#     distribution,
#     search_rate,
#     maintenance_rate,
#     assimilation_rate,
#     egg_mass,
#     spawning_interval,
#     vgscs,
#     damage_baseline,
#     repair_init,
#     repair_decay,
#     config
#     ):

#     config = read_config(f"config/parameters/tests/{config}.json")
#     config["agents"]["parameters"]["search_rate_max"] = search_rate
#     config["agents"]["parameters"]["assimilation_rate_max"] = assimilation_rate
#     config["agents"]["parameters"]["maintenance_rate_max"] = maintenance_rate / 1e5
#     config["agents"]["parameters"]["spawning_interval"] = spawning_interval
#     config["agents"]["parameters"]["egg_dry_mass"] = egg_mass
#     config["agents"]["parameters"]["distribution"] = distribution
#     config["agents"]["parameters"]["rate_damage_baseline"] = damage_baseline / 1e12
#     config["agents"]["parameters"]["rate_repair_decay"] = repair_decay / 1e12
#     config["agents"]["structures"]["VGSodiumChannels"]["volumetric_channel_concentration"] = vgscs * 1e29
#     config["agents"]["X0"]["Repair"]["rate"] = repair_init / 1e16
#     # create simulation
#     s = create_sim(config)
#     # run the simulation
#     s.set()
#     s.run()

#     data = s.experiment.get_tensor(return_time=True)[:, 0, :]
#     s.plot_life_history(show=True, store=False)

#     print(f"{'maximum size:': <20} {round(max(data[:, 2]), 2)} mm")
#     print(f"{'total offspring:': <20} {sum(np.nan_to_num(data[:, 3]))}")
#     print(f"{'maintenance rate:': <20} {maintenance_rate / 1e5}")
#     print(f"{'damage baseline:': <20} {damage_baseline / 1e12}")
#     print(f"{'repair init:': <20} {repair_init / 1e16}")
#     print(f"{'repair decay:': <20} {repair_decay / 1e12}")
#     print(f"{'VGSCs:': <20} {vgscs * 1e29}")
       
# # store the data
# # s.store_sim_results()
# # s.store_observations_csv()


# distribution=widgets.FloatSlider(min=0.05, max=0.95, step=0.01, value=0.5)
# search_rate=widgets.FloatSlider(min=1e-5, max=2e-3, value=7e-4, step=2e-5, readout_format=".5f")
# maintenance_rate=widgets.FloatSlider(min=1e-6, max=1e-3, step=1e-5, value=8e-3, readout_format=".4f")
# assimilation_rate=widgets.FloatSlider(min=1e-5, max=5e-2, step=5e-5, value=4e-5, readout_format=".4f")
# egg_mass=widgets.FloatSlider(min=0.001, max=.05, step=0.001, value=0.018, readout_format=".3f")
# spawning_interval=widgets.IntSlider(min=1e4, max=1e6, step=1e4, value=1.4e5)
# damage_baseline=widgets.IntSlider(min=0, max=100000, step=1, value=0)
# repair_decay=widgets.IntSlider(min=0, max=100000, step=100, value=0)
# vgscs=widgets.IntSlider(min=0, max=100, step=1, value=10)
# repair_init=widgets.FloatSlider(min=0.0, max=100, step=1, value=14)
# config=widgets.Dropdown(options=[
#     "daphnia_expo_experiment", 
#     "daphnia_control_experiment"
# ])


# interact(
#     run_sim, 
#     distribution=distribution, 
#     search_rate=search_rate,
#     assimilation_rate=assimilation_rate,
#     maintenance_rate=maintenance_rate, 
#     egg_mass=egg_mass,
#     spawning_interval=spawning_interval, 
#     damage_baseline=damage_baseline,
#     vgscs=vgscs,
#     repair_init=repair_init,
#     repair_decay=repair_decay,
#     config=config
# )



import time
import configparser
import os
import json
import numpy as np
import sys
import xarray as xr
from glob import glob
from importlib import import_module
from pymob.utils.misc import get_host
from pymob.utils.errors import errormsg

def create_fname_date_ver(dirname="./", fname="file", fextension=".txt"):
    ver = 0
    file = dirname + fname + time.strftime("%Y%m%d") + "_" + str(ver) + fextension
    while os.path.isfile(file):
        ver += 1
        file = dirname + fname + time.strftime("%Y%m%d") + "_" + str(ver) + fextension

    return file

def opt(a, b, c):
    """
    tests arguments for their state and returns with priority a, b and then c
    """
    return a if a is not None else b if b is not None else c

def is_number(s):
    try:
        float(s)
        return True
    except ValueError:
        return False


def read_config(config_file):
    # get last bit of file name after dot
    ext = config_file.split(".")[-1]

    if ext == "json":
        with open(config_file, "r") as f:
            return json.load(f)

    if ext == "cfg":
        converters={
            'list': lambda x: x if " " not in x else [i.strip() for i in x.split(' ')],
            'strint': lambda x: int(x) if x.isdigit() else x,
            'strfloat': lambda x: float(x) if is_number(x) else x,
            'listfloat': lambda x: x if " " not in x else [float(i.strip()) for i in x.split(' ')]
        }
        config = configparser.ConfigParser(converters=converters)
        _ = config.read(config_file)
        return config

def parse_config_section(section, method: str="strint"):
    """
    method can be one of 'strint' (mix of string and integers) or "list"
    if lists are also in the section
    """
    getter = getattr(section, f"get{method}")
    return {k:getter(k) for k in section.keys()}

def store_sbi_simulations(path, theta, x, simname=None):
    if simname == None:
        simname = time.strftime("%Y%m%d%H%M")
    simloc = os.path.normpath(os.path.join(path, simname))
    os.makedirs(simloc, exist_ok=True)
    np.savetxt(os.path.join(simloc, "theta.txt"), theta.numpy())
    np.savetxt(os.path.join(simloc, "x.txt"), x.numpy())

def read_settings():
    try:
        return read_config("config/settings.json")
    except FileNotFoundError:
        return read_config("config/settings_default.json")

def _insert_username(settings, dir):
    username = settings["username_remote"]
    return dir.replace("_USERNAME_", username)

def prepare_sbi(sbi_config_file):
    config = read_config(sbi_config_file)
    sim_conf = read_config(config["simulation_config"])

    # get the correct file
    settings = read_settings()
    host = get_host(settings["localhosts"])
    output_dir = os.path.join(config[f"output_{host}"])
    
    output = _adapt_output_to_machine(output_dir=output_dir)

    os.makedirs(config["plots"], exist_ok=True)

    return config, sim_conf, output

def unixify_path(path):
    normed_path_os = os.path.normpath(path)
    unix_path = normed_path_os.replace(os.path.sep, "/")
    return unix_path

def scenario_file(file, case_study, scenario, pkg_dir="case_studies"):
    file = os.path.join(pkg_dir, case_study, "scenarios", scenario, file)
    assert os.path.exists(file), f"{file} was not found"
    return file

def case_study_output(case_study, scenario, pkg_dir="case_studies"):
    study_name = os.path.basename(case_study)
    output = os.path.normpath(os.path.join(
        pkg_dir, study_name, "results", scenario))
    return pkg_dir, output

def prepare_scenario(case_study, scenario, input_files=[]):
    pkg_dir, output = case_study_output(case_study, scenario)
    
    paths_input_files = []
    for file in input_files:
        fp = scenario_file(file, case_study, scenario)
        paths_input_files.append(fp)

    output_path = reroute_output_to_base(output)

    return paths_input_files, output_path


def reroute_output_to_base(output):
    settings = read_settings()
    return unixify_path(os.path.join(
        settings[f"output_base"], 
        unixify_path(output)))

def prepare_casestudy(case_study, config_file, pkg_dir="case_studies"):
    # read sbi-config, sim-config and settings
    assert len(case_study) == 2, "case_study must be a list with first entry <case_study_directory> and 2nd entry the scenario name"
    config = read_config(scenario_file(config_file, *case_study, pkg_dir=pkg_dir))
    
    config["case-study"]["root"] = os.getcwd()
    config["case-study"]["name"] = case_study[0]
    config["case-study"]["scenario"] = case_study[1]

    # store package dir
    path_case_study = os.path.join(pkg_dir, case_study[0])
    config["case-study"]["package"] = path_case_study

    # append relevant paths to sys
    root = os.path.join(config.get("case-study", "root"), pkg_dir)
    if root not in sys.path:
        sys.path.append(root)
    
    package = os.path.join(config.get("case-study", "root"), config.get("case-study", "package"))
    if package not in sys.path:
        sys.path.append(package)

    return config

def prepare_casestudy_sbi(case_study):
    # read sbi-config, sim-config and settings
    config = read_config(scenario_file("sbi.json", *case_study))
    assert len(case_study) == 2, "case_study must be a list with first entry <case_study_directory> and 2nd entry the scenario name"
    sim_conf = prepare_scenario(*case_study)
    settings = read_settings()

    # determine directory depending of whether working local or remote
    host = get_host(settings["localhosts"])
    output_dir = unixify_path(os.path.join(
        settings[f"{host}_output_base"], 
        unixify_path(sim_conf["simulation"]["output"]).replace(
            "simulation", "sbi")
    ))

    # adapt output to the type of machine being worked on
    output = _adapt_output_to_machine(output_dir=output_dir)
    
    # create plot dir and pass to sbi-config
    plot_dir = unixify_path(sim_conf["simulation"]["output"]).replace(
        "simulation", "plots")
    os.makedirs(plot_dir, exist_ok=True)
    config["plots"] = plot_dir
    
    return config, sim_conf, output

def _adapt_output_to_machine(output_dir):
    settings = read_settings()
    host = get_host(settings["localhosts"])

    if host == "remote":
        output_dir = os.path.normpath(_insert_username(settings, output_dir))

        try:
            os.makedirs(output_dir, exist_ok=True)
        except PermissionError:
            if os.path.isabs(output_dir):
                output_dir_old = output_dir
                output_dir = os.path.abspath(output_dir_old[1:])  # remove leading /
                sys.stdout.write(
                    f"no permission to write to {output_dir_old} using {output_dir}"
                )
                os.makedirs(output_dir, exist_ok=True)
    
    return unixify_path(output_dir)

def import_package(package_path):
    """
    this script handles the import of a case study without the typical 
    __init__.py file. It iterates through all .py files in the root directory
    of the case study (typically: sim, mod, stats, plot, data, prior)
    and imports them with ()
    package     path to package from project dir (e.g. "model/core/daphnia")
    """
    package = os.path.basename(package_path)
    package_dir = os.path.dirname(package_path)
    sys.path.append(package_dir)

    # it is for me only possible to import the specific modules and not 
    # only the package and then refer to the modules
    for mod in glob(os.path.join(package_dir, package, "*.py")):
        module = os.path.basename(mod).split('.')[0]
        try:
            _ = import_module(f"{package}.{module}")
        except ModuleNotFoundError as e:
            raise ModuleNotFoundError(errormsg(
                """The module could not be found. This can happen if the name
                of the case-study conflicts with another package. To avoid this
                conflict, give the case study a more unique name. E.g.
                case_study_test. You can confirm, whether another module with that name
                exists, by opening python in the command line and typing the
                name of your case study, if there is no error, it means that the
                package was found in your environment. You can then safely assume
                that the error came from a naming conflict and resolve it as suggested.
                """
            ))

    pkg = import_module(package)
    return pkg

def sequential_filename_iterator(path, sep="_"):
    """
    file storing function that appends a running number to the 
    filename, depending on the number of files in the directory. Note that this
    does only work if files are created sequentially. It will probably produce 
    errors if the files are produced in parallel.
    """
    # split path
    directory, file = os.path.split(path)

    # split filename
    filename, fileext = file.split(".")

    # find number of files in directory
    nfiles = len([
        f for f in os.listdir(directory) 
        if os.path.isfile(os.path.join(directory, f))
    ])
    filename = filename + sep + str(nfiles).zfill(5)

    newpath = os.path.join(directory, filename + "." + fileext)

    return newpath

def serialize(ds, convert_time=True):
    """
    function serializes a dataset. By default converts a time coordinate from
    datetime64[ns] to float[ns]
    """

    if convert_time:
        ds["time"] = ds.time.astype(float)

    return json.dumps(ds.to_dict())

def deserialize(dct, convert_time=True):
    ds = xr.Dataset.from_dict(json.loads(dct))

    if convert_time:
        ds["time"] = ds.time.astype("timedelta64[ns]")

    return ds

def unnest(d, flat, parent_key="", sep="."):
    for child_key, value in d.items():
        if parent_key != "":
            key = f"{parent_key}{sep}{child_key}"
        else:
            key = child_key
            
        if isinstance(value, dict):
            flat = unnest(d=value, flat=flat, parent_key=key)
        else:            
            flat.append((key, value))
    
    return flat


import inspect

class SolverBase:
    """
    The idea of creating a solver as a class is that it is easier
    to pass on important arguments of the simulation relevant to the 
    Solver. Therefore a solver can access all attributes of an Evaluator
    """
    def __call__(self, **kwargs):
        return self.solve(**kwargs)
    
    @staticmethod
    def solve():
        raise NotImplementedError("Solver must implement a solve method.")


def mappar(func, parameters, exclude=[]):
    func_signature = inspect.signature(func).parameters.keys()
    model_param_signature = [p for p in func_signature if p not in exclude]
    model_args = [parameters.get(k) for k in model_param_signature]

    return tuple(model_args)



import numpy as np
import argparse
from pymob.sims.optimization import IndyOptimizer
from pymob.sims.simulation import create_sim, update_parameters
from pymob.utils.store_file import read_config

parser = argparse.ArgumentParser(
    prog="Parameter optimization on observation timeseries of individuals")
parser.add_argument("-s", "--sim_conf", type=str,
                    default="config/parameters/optimization_indy.json", 
                    help="parameter file for the simulation")
parser.add_argument("-o", "--opt_conf", type=str,
                    default="config/parameters/opt_conf_indy_debug.json",
                    help="(hyper)parameter file for the optimization process")
parser.add_argument("-i", "--iterations", type=int, default=0, 
                    help="set iterations of basinhopping optimizer - defaults \
                          to iterations of opt_conf if not set.")
parser.add_argument("-t", "--temperature", type=float, default=0,
                    help="The “temperature” parameter for the accept or reject \
                          criterion. Higher “temperatures” mean that larger jumps\
                          in function value will be accepted. For best results T\
                          should be comparable to the separation (in function value)\
                          between local minima.")
parser.add_argument("-p", "--step", type=float, default=0,
                    help="maximum stepsize used in random displacement")

args = parser.parse_args()


def loss_function(s, use_sim_cols, exp_data, scaler):
    sim_data = s.events.get_tensor()[:, 0, use_sim_cols]

    # calculate the difference between simulation results and
    # experiment results and remove nan values (these result only because
    # of zero values)
    diff = scaler.transform(np.nan_to_num(sim_data, nan=0)) - exp_data

    ssq=np.sum(np.nan_to_num(diff, nan=0)**2)
    return ssq

def iteration(x, config, scales, use_sim_cols, exp_data, scaler, parnames):
    config = update_parameters(config, x, parnames, scales)
    s = create_sim(config)
    s.run(progress_bar=False)
    loss = loss_function(s, use_sim_cols, exp_data, scaler)
    # print("memory:", asizeof.asizeof(s))
    print("loss:", loss)
    return loss

sim_conf = args.sim_conf
opt_conf = read_config(args.opt_conf)

if args.iterations > 0:
    opt_conf["iterations"] = args.iterations

if args.temperature > 0:
    opt_conf["temperature"] = args.temperature

if args.step > 0:
    opt_conf["step"] = args.step

print("using optimization configuration:\n", opt_conf)

io = IndyOptimizer(
    f=iteration, 
    params=opt_conf["parameters"], 
    temperature=opt_conf["temperature"],
    stepsize=opt_conf["step"],
    niter=opt_conf["iterations"],
    interval=opt_conf["interval"],
    sim_config=sim_conf,
    seed=opt_conf["seed"],
    minimizer_kwargs=opt_conf["minimizer_kwargs"],
    optimize_on_sample=opt_conf["sample"],
    optimize_on_features=opt_conf["optimize_on"])

io.optim()
io.plot_results(update_parameters(io.sim_config, io.result.x,
                                  io.parnames, io.parameter_scales))
io.store_data()
# there are still some problems some debugging loops are necessary
# [x] standardize data and results
# [x] make death deterministic
# [x] store results
# [ ] write wrapper for optimization script for calibration on cluster



import numpy as np
from matplotlib import pyplot as plt
from matplotlib import ticker
from pymob.utils.misc import Date2Delta
import arviz as az

def plot_loghist(x, name="", bins=10, ax=None, hdi=False, decorate=True, 
                 color="tab:blue", alpha=1, legend=None):
    _, bins = np.histogram(x, bins=bins)
    median = np.median(x)
    if hdi:
        assert hasattr(x, "chain"), "chain must be in coordinates, if hdi=True"
        assert hasattr(x, "draw"), "draw must be in coordinates, if hdi=True"
        lq, uq = az.hdi(x.unstack(), hdi_prob=.95).x.values
    else:
        lq, uq = np.quantile(x, [0.025, 0.975])
    logbins = np.logspace(np.log10(bins[0]), np.log10(bins[-1]), len(bins))
    if ax is None:
        ax = plt.subplot(111)
    lhist, lbins, _ = ax.hist(x, bins=logbins, color=color, alpha=alpha, label=legend)

    if decorate:
        plot_kwargs = dict(rotation=90, va="bottom", ha="center")
        bbox=dict(facecolor='white', alpha=0.8, linewidth=0)
        ax.vlines([lq, median, uq], 0, np.max(lhist)*1.1, linestyle="--", 
                    color = ["grey", "black", "grey"])
        ax.text(lq, np.max(lhist)*0.3, f"{lq:.1e}", color="black", bbox=bbox, **plot_kwargs)
        ax.text(uq, np.max(lhist)*0.3, f"{uq:.1e}", color="black", bbox=bbox, **plot_kwargs)
        ax.text(median, np.max(lhist)*0.3, f"{median:.1e}", bbox=bbox, **plot_kwargs)
        ax.text(0.05, 0.95, name, transform=ax.transAxes, ha="left", va="top")

    ax.set_xscale("log")
    ax.set_ylim(0, np.max(lhist)*1.1)
    return ax
    # plt.savefig("work/case_studies/core_daphnia/results/expo_control_3/plots/test.png")

def plot_hist(x, name="", bins=10, ax=None, hdi=False, decorate=True):
    median = np.median(x)
    if hdi:
        assert hasattr(x, "chain"), "chain must be in coordinates, if hdi=True"
        assert hasattr(x, "draw"), "draw must be in coordinates, if hdi=True"
        lq, uq = az.hdi(x.unstack(), hdi_prob=.95).x.values
    else:
        lq, uq = np.quantile(x, [0.025, 0.975])
    if ax is None:
        ax = plt.subplot(111)
    lhist, lbins, _ = ax.hist(x, bins=bins)

    if decorate:
        plot_kwargs = dict(rotation=90, va="bottom", ha="center")
        bbox=dict(facecolor='white', alpha=0.5, edgecolor=None)
        ax.vlines([lq, median, uq], 0, np.max(lhist)*1.1, linestyle="--", 
                    color = ["grey", "black", "grey"])
        ax.text(lq, np.max(lhist)*0.3, f"{lq:.1e}", color="black", bbox=bbox, **plot_kwargs)
        ax.text(uq, np.max(lhist)*0.3, f"{uq:.1e}", color="black", bbox=bbox, **plot_kwargs)
        ax.text(median, np.max(lhist)*0.3, f"{median:.1e}", bbox=bbox, **plot_kwargs)
        ax.text(np.min(lbins), np.max(lhist)*0.9, name)
    
    ax.set_ylim(0, np.max(lhist)*1.1)

def combine_legends(axes):
    handles = []
    labels = []

    for ax in axes:
        h, l = ax.get_legend_handles_labels()
        handles.extend(h)
        labels.extend(l)

    return handles, labels

def _format_x(origin, ax):
    format_func = Date2Delta(origin=origin)
    formatter = ticker.FuncFormatter(format_func)
    ax.xaxis.set_major_formatter(formatter)



import os
from functools import lru_cache
import tempfile

import numpy as np
import pyabc
import xarray as xr
import arviz as az
from matplotlib import pyplot as plt
from pathos import multiprocessing as mp

from pymob.simulation import SimulationBase


class PyabcBackend:
    def __init__(
        self, 
        simulation: SimulationBase
    ):
        self.simulation = simulation
        self.evaluator = self.model_parser()
        self.prior = self.prior_parser(simulation.free_model_parameters)
        self.distance_function = self.distance_function_parser()
        self.observations = simulation.observations

        self.abc = None
        self.history = None
        self.posterior = None

    @property
    def sampler(self):
        return self.simulation.config.get("inference.pyabc", "sampler")
    
    @property
    def population_size(self):
        return self.simulation.config.getint("inference.pyabc", "population_size")
    
    @property
    def minimum_epsilon(self):
        return self.simulation.config.getfloat("inference.pyabc", "minimum_epsilon")
    
    @property
    def min_eps_diff(self):
        return self.simulation.config.getfloat("inference.pyabc", "min_eps_diff")
    
    @property
    def max_nr_populations(self):
        return self.simulation.config.getint("inference.pyabc", "max_nr_populations")
    
    @property
    def database(self):
        tmp = tempfile.gettempdir()
        dbp = self.simulation.config.get(
            "inference.pyabc", "database_path", fallback=f"{tmp}/pyabc.db"
        )
        if os.path.isabs(dbp):
            return dbp
        else:
            return os.path.join(self.simulation.output_path, dbp)
    
    @property
    def redis_password(self):
        return self.simulation.config.get("inference.pyabc.redis", "password", "nopassword")
    
    @property
    def redis_port(self):
        return self.simulation.config.getint("inference.pyabc.redis", "port", 1111)
    
    @property
    def history_id(self):
        return self.simulation.config.getint(
            "inference.pyabc", "eval.history_id", fallback=-1
        )

    @property
    def model_id(self):
        return self.simulation.config.getint(
            "inference.pyabc", "eval.model_id", fallback=0
        )
    
    @property
    def n_predictions(self):
        return self.simulation.config.getint(
            "inference.pyabc", "eval.n_predictions", fallback=50
        )
    
    @staticmethod
    def param_to_prior(par):
        parname = par.name
        distribution, cluttered_arguments = par.prior.split("(", 1)
        param_strings = cluttered_arguments.split(")", 1)[0].split(",")
        params = {}
        for parstr in param_strings:
            key, val = parstr.split("=")
            params.update({key:float(val)})

        return parname, distribution, params

    @classmethod
    def prior_parser(cls, free_model_parameters: list):

        prior_dict = {}
        for mp in free_model_parameters:
            name, distribution, params = cls.param_to_prior(par=mp)

            prior = pyabc.RV(distribution, **params)
            prior_dict.update({name: prior})

        return pyabc.Distribution(**prior_dict)

    @property
    def plot_function(self):
        return self.simulation.config.get(
            "inference.pyabc", "plot_function", 
            fallback=None
        )

    def plot(self):
        plot_func = getattr(self.simulation, self.plot_function)
        plot_func()
        

    def model_parser(self):
        def model(theta):
            evaluator = self.simulation.dispatch(theta=dict(theta))
            evaluator()
            return {k: np.array(v) for k, v in evaluator.Y.items()}
        return model
    
    def distance_function_parser(self):
        def distance_function(x, x0):
            Y = {k: v for k, v in x.items() if k in self.simulation.data_variables}
            obj_name, obj_value = self.simulation.objective_function(results=Y)
            return obj_value
        
        return distance_function

    def run(self):
        n_cores = self.simulation.n_cores
        print(f"Using {n_cores} CPU cores", flush=True)

        # before launch server in bash with `redis-server --port 1803`
        if self.sampler.lower() == "RedisEvalParallelSampler".lower():
            abc_sampler = pyabc.sampler.RedisEvalParallelSampler(
                host="localhost", 
                password=self.redis_password, 
                port=self.redis_port
            )

        elif self.sampler.lower() == "SingleCoreSampler".lower():
            abc_sampler = pyabc.sampler.SingleCoreSampler()

        elif self.sampler.lower() == "MulticoreParticleParallelSampler".lower():
            abc_sampler = pyabc.sampler.MulticoreParticleParallelSampler(
                n_procs=n_cores
            )

        elif self.sampler.lower() == "MulticoreEvalParallelSampler".lower():
            abc_sampler = pyabc.sampler.MulticoreEvalParallelSampler(
                n_procs=n_cores
            )

        else:
            raise NotImplementedError(
                "Sampler is not implemented. Choose one of: 'RedisEvalParallelSampler', " +
                "'SingleCoreSampler'"
            )


        self.abc = pyabc.ABCSMC(
            models=self.evaluator, 
            parameter_priors=self.prior, 
            distance_function=self.distance_function, 
            sampler=abc_sampler,
            population_size=self.population_size
        )

        self.history = self.abc.new("sqlite:///" + self.database)

        self.abc.run(
            minimum_epsilon=self.minimum_epsilon,
            min_eps_diff=self.min_eps_diff,
            max_nr_populations=self.max_nr_populations
        )


    def load_results(self):
        if self.history is None:
            self.history = pyabc.History(f"sqlite:///" + self.database)
        
        # set history id
        db_id = self.history_id
        self.history.id = self.history._find_latest_id() if db_id == -1 else db_id
        
        mod_id = self.model_id
        samples, w = self.history.get_distribution(m=mod_id, t=self.history.max_t)
        
        # re-sort parameters based on prior order
        samples = samples[self.prior.keys()]

        posterior = xr.DataArray(
            samples.values.reshape((1,*samples.values.shape)),
            coords=dict(
                chain=[1],
                draw=range(len(samples)),
                parameter=list(samples.columns)
            )
        )
        
        idata = az.from_dict(
            posterior={key: col.values for key, col in samples.items()},
            dims=self.posterior_data_structure,
            coords=self.posterior_coordinates
        )
        self.idata = idata
        # posterior
        self.posterior = Posterior(posterior)

    @property
    def posterior_data_structure(self):
        data_structure = self.simulation.data_structure.copy()
        data_structure_loglik = {f"{dv}_obs": dims for dv, dims in data_structure.items()}
        data_structure.update(data_structure_loglik)
        return data_structure

    @property
    def posterior_coordinates(self):
        posterior_coords = self.simulation.coordinates.copy()
        posterior_coords.update({
            "draw": list(range(self.population_size)), 
            "chain": [0]
        })
        return posterior_coords

    def plot_chains(self):

        distributions = []
        for t in range(self.history.max_t + 1):
        # for t in range(2):
            print(f"iteration: {t}", end="\r")
            par_values, _ = self.history.get_distribution(m=0, t=t)
            post = par_values.to_xarray().to_array()
            post["id"] = range(len(post.id))
            post = post.assign_coords(iteration=t)
            distributions.append(post)

        trajectory = xr.concat(distributions, dim="iteration")
        
        parameters = trajectory.coords["variable"].values
        nit = len(trajectory.iteration)
        color="tab:blue"
        Np = len(parameters)
        fig, axes = plt.subplots(nrows=Np, ncols=2, figsize=(10, 2*Np),
            gridspec_kw=dict(width_ratios=[1, 1]))
        
        if len(axes.shape) == 1:
            axes = axes.reshape((Np, 2))
        
        for i, par in enumerate(parameters):
            ax = axes[i, 0]

            ax.plot(trajectory.iteration, trajectory.sel(variable=par), 
                color=color,
                # s=5,
                alpha=.01)
                
            # takes the last 
            ax2 = axes[i, 1]
            ax2.hist(trajectory.sel(variable=par, iteration=nit-1), 
                    alpha=.85, color=color)
            ax2.set_ylabel(f"{par.split('.')[-1].replace('_', ' ')}")
            
            ax.set_xlabel("iteration")
            ax.set_yscale("log")
            ax.set_ylabel(f"{par.split('.')[-1].replace('_', ' ')}")

        fig.subplots_adjust(wspace=.3, hspace=.4)

        return fig

    def store_results(self):
        """results are stored by default in database"""
        pass
   
    @lru_cache
    def posterior_predictions(self, n=50, seed=1):
        rng = np.random.default_rng(seed)
        post = self.posterior
        total_samples = post.samples.shape[1]

        # draw samples from posterior
        posterior_samples = rng.choice(post.samples.draw, size=n, replace=False)

        def predict(posterior_sample_id):
            params = post.draw(i=posterior_sample_id)
            evaluator = self.simulation.dispatch(params.to_dict())
            evaluator()
            res = evaluator.results
            res = res.assign_coords({"draw": posterior_sample_id, "chain": 1})
            res["params"] = params.samples
            res = res.expand_dims(("chain", "draw"))
            return res
        
        print(f"Using {self.simulation.n_cores} CPUs")
        if self.simulation.n_cores == 1:
            results = list(map(predict, posterior_samples))
        else:
            with mp.ProcessingPool(self.simulation.n_cores) as pool:        
                results = pool.map(predict, posterior_samples)
            
        return xr.combine_by_coords(results)

    def plot_predictions(
            self, data_variable: str, x_dim: str, ax=None, subset={}
        ):
        obs = self.simulation.observations.sel(subset)
        
        post_pred = self.posterior_predictions(
            n=self.n_predictions, 
            # seed only controls the parameters samples drawn from posterior
            seed=self.simulation.seed
        ).sel(subset)

        hdi = az.hdi(post_pred, .95)

        if ax is None:
            ax = plt.subplot(111)
        
        y_mean = post_pred[data_variable].mean(dim=("chain", "draw"))
        ax.plot(
            post_pred[x_dim].values, y_mean.values, 
            color="black", lw=.8
        )

        ax.fill_between(
            post_pred[x_dim].values, *hdi[data_variable].values.T, 
            alpha=.5, color="grey"
        )

        ax.plot(
            obs[x_dim].values, obs[data_variable].values, 
            marker="o", ls="", ms=3
        )
        
        ax.set_ylabel(data_variable)
        ax.set_xlabel(x_dim)

        return ax

class Posterior:
    def __init__(self, samples):
        self.samples = samples

    def __repr__(self):
        return str(self.samples)

    def to_dict(self):
        theta = self.samples
        return {par:float(theta.sel(parameter=par)) 
            for par in theta.parameter.values}

    def draw(self, i):
        return Posterior(self.samples.sel(draw=i, chain=1))

    def mean(self):
        return Posterior(self.samples.mean(dim=("chain", "draw")))





import os
import warnings
import configparser
from functools import partial
import multiprocessing as mp
from typing import Callable, Dict
from multiprocessing.pool import ThreadPool, Pool

import numpy as np
import xarray as xr
import dpath.util as dp
from sklearn.preprocessing import StandardScaler, MinMaxScaler
from toopy import Param, FloatParam, IntParam
from toopy import benchmark

from pymob.utils.errors import errormsg
from pymob.utils.store_file import scenario_file, parse_config_section
from pymob.sim.evaluator import Evaluator, create_dataset_from_dict, create_dataset_from_numpy

def update_parameters_dict(config, x, parnames):
    for par, val, in zip(parnames, x):
        key_exist = dp.set(config, glob=par, value=val, separator=".")
        if key_exist != 1:
            raise KeyError(
                f"prior parameter name: {par} was not found in config. " + 
                f"make sure parameter name was spelled correctly"
            )
    return config



class SimulationBase:
    model: Callable
    def __init__(
        self, 
        config: configparser.ConfigParser, 
    ) -> None:
        
        self.config = config
        self.model_parameters: Dict = {}
        self.observations = None
        self._objective_names = []
        self._seed_buffer_size = self.n_cores * 2
        
        # seed gloabal RNG
        self.RNG = np.random.default_rng(self.seed)
        # draw a selection of 1e8 integers for using those throughout the
        # simulation
        self._random_integers = self.create_random_integers(n=self._seed_buffer_size)

        self.initialize(input=self.input_file_paths)
        
        if self.observations is not None:
            self.create_data_scaler()
        
        coords = self.set_coordinates(input=self.input_file_paths)
        self.coordinates = self.create_coordinates(coordinate_data=coords)
        self.var_dim_mapper = self.create_dim_index()
        self.free_model_parameters  = self.set_free_model_parameters()

        if not os.path.exists(self.output_path):
            os.makedirs(self.output_path)

        self.validate()
        # TODO: set up logger
        self.parameterize = partial(self.parameterize, model_parameters=self.model_parameters)


    def __repr__(self) -> str:
        return f"Simulation(case_study={self.case_study}, scenario={self.scenario})"

    def print_config(self):
        print("Simulation configuration", end="\n")
        print("========================")
        for section in self.config.sections():
            print("\n")
            print(f"[{section}]", end="\n")

            for opt, val in self.config.items(section):
                print(f"{opt} = {val}", end="\n")

        print("========================", end="\n")

    def benchmark(self, n=100):
        evaluator = self.dispatch(theta=self.model_parameter_dict)
        evaluator() 

        @benchmark
        def run_bench():
            for i in range(n):
                evaluator()
        
        print(f"\nBenchmarking with {n} evaluations")
        print(f"=================================")
        run_bench()
        print(f"=================================\n")
        

        
    def dispatch(self, theta, **evaluator_kwargs):
        """Dispatch an evaluator, which will compute the model at parameters
        (theta). Evaluators are advantageous, because they are easier serialized
        than the whole simulation object. Comparison can then happen back in 
        the simulation.

        Theoretically, this could also be used to constrain coordinates etc, 
        before evaluating.  
        """
        model_parameters = self.parameterize(theta)
        
        # TODO: make sure the evaluator has all arguments required for solving
        # model
        # TODO: Check if model is bound. If yes extract
        if hasattr(self.solver, "__func__"):
            solver = self.solver.__func__
        else:
            solver = self.solver

        if hasattr(self.model, "__func__"):
            model = self.model.__func__
        else:
            model = self.model

        stochastic = self.config.get("simulation", "modeltype", fallback=False)
            
        evaluator = Evaluator(
            model=model,
            solver=solver,
            parameters=model_parameters,
            dimensions=self.dimensions,
            var_dim_mapper=self.var_dim_mapper,
            data_structure=self.data_structure,
            data_variables=self.data_variables,
            coordinates=self.coordinates,
            # TODO: pass the whole simulation settings section
            stochastic=True if stochastic == "stochastic" else False,
            **evaluator_kwargs
        )

        return evaluator

    def evaluate(self, theta):
        """Wrapper around run to modify paramters of the model.
        """
        self.model_parameters = self.parameterize(theta)
        return self.run()
    
    def compute(self):
        """
        A wrapper around run, which catches errors, logs, does post processing
        """
        warnings.warn("Discouraged to use self.Y constructs. Instability suspected.", DeprecationWarning, 2)
        self.Y = self.evaluate(theta=self.model_parameter_dict)

    def interactive(self):
        import ipywidgets as widgets
        from IPython.display import display, clear_output
        
        def interactive_output(func, controls):
            out = widgets.Output(layout={'border': '1px solid black'})
            def observer(change):
                theta={key:s.value for key, s in sliders.items()}
                widgets.interaction.show_inline_matplotlib_plots()
                with out:
                    clear_output(wait=True)
                    func(theta)
                    widgets.interaction.show_inline_matplotlib_plots()
            for k, slider in controls.items():
                slider.observe(observer, "value")
            widgets.interaction.show_inline_matplotlib_plots()
            observer(None)
            return out

        sliders = {}
        for par in self.free_model_parameters:
            s = widgets.FloatSlider(
                par.value, description=par.name, min=par.min, max=par.max,
                step=par.step
            )
            sliders.update({par.name: s})

        def func(theta):
            evaluator = self.dispatch(theta=theta)
            evaluator()
            self.plot(results=evaluator.results)

        out = interactive_output(func=func, controls=sliders)

        display(widgets.HBox([widgets.VBox([s for _, s in sliders.items()]), out]))
    
    def set_inferer(self, backend):
        if backend == "pyabc":
            from pymob.inference.pyabc_backend import PyabcBackend

            self.inferer = PyabcBackend(simulation=self)

        elif backend == "pymoo":
            from pymob.inference.pymoo_backend import PymooBackend

            self.inferer = PymooBackend(simulation=self)

        elif backend == "numpyro":
            from pymob.inference.numpyro_backend import NumpyroBackend

            self.inferer = NumpyroBackend(simulation=self)
    
        else:
            raise NotImplementedError(f"Backend: {backend} is not implemented.")

    def check_dimensions(self, dataset):
        """Check if dataset dimensions match the specified dimensions.
        TODO: Name datasets for referencing them in errormessages
        """
        ds_dims = list(dataset.dims.keys())
        in_dims = [k in self.dimensions for k in ds_dims]
        assert all(in_dims), IndexError(
            "Not all dataset dimensions, were not found in specified dimensions. "
            f"Settings(dims={self.dimensions}) != dataset(dims={ds_dims})"
        )
        
    def dataset_to_2Darray(self, dataset: xr.Dataset): 
        self.check_dimensions(dataset=dataset)
        array_2D = dataset.stack(multiindex=self.dimensions)
        return array_2D.to_array().transpose("multiindex", "variable")

    def array2D_to_dataset(self, dataarray: xr.DataArray): 
        dataset_2D = dataarray.to_dataset(dim="variable")      
        return dataset_2D.unstack().transpose(*self.dimensions)

    def create_data_scaler(self):
        """Creates a scaler for the data variables of the dataset over all
        remaining dimensions.
        In addition produces a scaled copy of the observations
        """
        # make sure the dataset follows the order of variables specified in
        # the config file. This is important so also in the simulation results
        # the scalers are matched.
        ordered_dataset = self.observations[self.data_variables]
        obs_2D_array = self.dataset_to_2Darray(dataset=ordered_dataset)
        # scaler = StandardScaler()
        scaler = MinMaxScaler()
        
        # add bounds to array of observations and fit scaler
        lower_bounds, upper_bounds = self.data_variable_bounds
        stacked_array = np.row_stack([lower_bounds, upper_bounds, obs_2D_array])
        scaler.fit(stacked_array)

        self.scaler = scaler
        self.print_scaling_info()

        scaled_obs = self.scale_(self.observations)
        self.observations_scaled = scaled_obs

    def print_scaling_info(self):
        scaler = type(self.scaler).__name__
        for i, var in enumerate(self.data_variables):
            print(
                f"{scaler}(variable={var}, "
                f"min={self.scaler.data_min_[i]}, max={self.scaler.data_max_[i]})"
            )

    def scale_(self, dataset: xr.Dataset):
        ordered_dataset = dataset[self.data_variables]
        data_2D_array = self.dataset_to_2Darray(dataset=ordered_dataset)
        obs_2D_array_scaled = data_2D_array.copy() 
        obs_2D_array_scaled.values = self.scaler.transform(data_2D_array)
        return self.array2D_to_dataset(obs_2D_array_scaled)

    @property
    def results(self):
        warnings.warn("Discouraged to use results property.", DeprecationWarning, 2)
        return self.create_dataset_from_numpy(
            Y=self.Y, 
            Y_names=self.data_variables, 
            coordinates=self.coordinates
        )

    def results_to_df(self, results):
        if isinstance(results, xr.Dataset):
            return results
        elif isinstance(results, dict):
            return create_dataset_from_dict(
                Y=results, 
                coordinates=self.coordinates,
                data_structure=self.data_structure,
            )
        elif isinstance(results, np.ndarray):
            return create_dataset_from_numpy(
                Y=results,
                Y_names=self.data_variables,
                coordinates=self.coordinates,
            )
        else:
            raise NotImplementedError(
                "Results returned by the solver must be of type Dict or np.ndarray."
            )
    

    @property
    def results_scaled(self):
        scaled_results = self.scale_(self.results)
        # self.check_scaled_results_feasibility(scaled_results)
        return scaled_results

    def scale_results(self, Y):
        ds = self.create_dataset_from_numpy(
            Y=Y, 
            Y_names=self.data_variables, 
            coordinates=self.coordinates
        )
        return self.scale_(ds)

    def check_scaled_results_feasibility(self, scaled_results):
        """Parameter inference or optimization over many variables can only succeed
        in reasonable time if the results that should be compared are on approximately
        equal scales. The Simulation class, automatically estimates the scales
        of result variables, when observations are provided. 

        Problems can occurr when observations are on very narrow ranges, but the 
        simulation results can take much larger or lower values for that variable.
        As a result the inference procedure will almost exlusively focus on the
        optimization of this variable, because it provides the maximal return.

        The function warns the user, if simulation results largely deviate from 
        the scaled minima or maxima of the observations. In this case manual 
        minima and maxima should be given
        """
        max_scaled = scaled_results.max()
        min_scaled = scaled_results.min()
        if isinstance(self.scaler, MinMaxScaler):
            for varkey, varval in max_scaled.variables.items():
                if varval > 2:
                    warnings.warn(
                        f"Scaled results for '{varkey}' are {float(varval.values)} "
                        "above the ideal maximum of 1. "
                        "You should specify explicit bounds for the results variable."
                    )

            for varkey, varval in min_scaled.variables.items():
                if varval < -1:
                    warnings.warn(
                        f"Scaled results for '{varkey}' are {float(varval.values)} "
                        "below the ideal minimum of 0. "
                        "You should specify explicit bounds for the results variable."
                    )

    def validate(self):
        # TODO: run checks if the simulation was set up correctly
        #       - do observation dimensions match the model output (run a mini
        #         simulation with reduced coordinates to verify)
        #       - 
        pass

    @staticmethod
    def parameterize(free_parameters: list[Param], model_parameters) -> dict:
        """
        Optional. Set parameters and initial values of the model. 
        Must return a dictionary with the keys 'y0' and 'parameters'
        
        Can be used to define parameters directly in the script or from a 
        parameter file.

        Arguments
        ---------

        input: List[str] file paths of parameter/input files
        theta: List[Param] a list of Parameters. By default the parameters
            specified in the settings.cfg are used in this list. 

        returns
        -------

        tulpe: tuple of parameters, can have any length.
        """
        parameters = model_parameters["parameters"]
        y0 = model_parameters["y0"]

        parameters.update({p.name: p.value for p in free_parameters})
        return {"y0": y0, "parameters": parameters} 

    def run(self):
        """
        Implementation of the forward simulation of the model. Needs to return
        X and Y

        returns
        -------

        X: np.ndarray | xr.DataArray
        Y: np.ndarray | xr.DataArray
        """
        raise NotImplementedError
    
    def objective_function(self, results, **kwargs):
        func = getattr(self, self.objective)
        obj = func(results, **kwargs)

        if obj.ndim == 0:
            obj_value = float(obj)
            obj_name = "objective"
        if obj.ndim == 1:
            obj_value = obj.values
            obj_name = list(obj.coords["variable"].values)

        if len(self._objective_names) == 0:
            self._objective_names = obj_name

        return obj_name, obj_value

    def total_average(self, results):
        """objective function returning the total MSE of the entire dataset"""
        
        diff = (self.scale_(self.results_to_df(results)) - self.observations_scaled).to_array()
        return (diff ** 2).mean()

    def prior(self):
        raise NotImplementedError

    def initialize(self, input):
        """
        initializes the simulation. Performs any extra work, not done in 
        parameterize or set_coordinates. 
        """
        pass
    
    def dump(self):
        pass
        
    
    def plot(self):
        pass

    def create_coordinates(self, coordinate_data):
        if not isinstance(coordinate_data, tuple):
            coordinate_data = (coordinate_data, )

        assert len(self.dimensions) == len(coordinate_data), errormsg(
            f"""number of dimensions, specified in the configuration file
            must match the coordinate data (X) returned by the `run` method.
            """
        )

        coords = {dim: x_i for dim, x_i in zip(self.dimensions, coordinate_data)}
        return coords

    @staticmethod
    def create_dataset_from_numpy(Y, Y_names, coordinates):
        warnings.warn(
            "Use `create_dataset_from_numpy` defined in sim.evaluator",
            category=DeprecationWarning
        )
        n_vars = Y.shape[-1]
        n_dims = len(Y.shape)
        assert n_vars == len(Y_names), errormsg(
            """The number of datasets must be the same as the specified number
            of data variables declared in the `settings.cfg` file.
            """
        )

        # transpose Y to put the variable dimension first, then add the
        # remaining dimensions in order
        Y_transposed = Y.transpose((n_dims - 1, *range(n_dims - 1)))

        data_arrays = []
        for y, y_name in zip(Y_transposed, Y_names):
            da = xr.DataArray(y, coords=coordinates, name=y_name)
            data_arrays.append(da)

        dataset = xr.merge(data_arrays)

        return dataset

    @staticmethod
    def option_as_list(opt):
        if not isinstance(opt, (list, tuple)):
            opt_list = [opt]
        else:
            opt_list = opt

        return opt_list

    @property
    def input_file_paths(self):
        paths_input_files = []
        for file in self.input_files:
            fp = scenario_file(file, self.case_study, self.scenario)
            paths_input_files.append(fp)

        observation_files = self.config.getlist("case-study", "observations", fallback=[])
        if isinstance(observation_files, str):
            observation_files = [observation_files]

        for file in observation_files:
            if not os.path.isabs(file):
                fp = os.path.join(self.data_path, file)
            else:
                fp = file
            paths_input_files.append(fp)

        return paths_input_files

    # config as properties
    @property
    def dimensions(self):
        dims = self.config.getlist("simulation", "dimensions")
        return self.option_as_list(dims)

    @property
    def data_variables(self):
        data_vars = self.config.getlist("simulation", "data_variables")
        return self.option_as_list(data_vars)

    @property
    def input_files(self):
        input_files = self.config.getlist("simulation", "input_files", fallback=[])
        return self.option_as_list(input_files)
  
    @property
    def case_study_path(self):
        return self.config.get("case-study", "package")

    @property
    def root_path(self):
        return self.config.get("case-study", "root")

    @property
    def case_study(self):
        return self.config.get("case-study", "name")

    @property
    def scenario(self):
        return self.config.get("case-study", "scenario")

    @property
    def model_parameter_values(self):
        return [p.value for p in self.free_model_parameters]
    
    @property
    def model_parameter_names(self):
        return [p.name for p in self.free_model_parameters]
    
    @property
    def n_free_parameters(self):
        return len(self.free_model_parameters)

    @property
    def model_parameter_dict(self):
        return {p.name:p.value for p in self.free_model_parameters}


    @property
    def output_path(self):
        output_path_ = self.config.get("case-study", "output")
        if not os.path.isabs(output_path_):
            return os.path.join(
                os.path.relpath(output_path_),
                os.path.relpath(self.case_study_path), 
                "results",
                self.scenario,
            )
        else:
            return os.path.abspath(output_path_)

    @property
    def data_path(self):
        data_path_ = self.config.get("case-study", "data")
        if not os.path.isabs(data_path_):
            return os.path.join(
                self.case_study_path, 
                os.path.relpath(data_path_)
            )
        else:
            return os.path.abspath(data_path_)
       

    @property
    def data_variable_bounds(self):
        lower_bounds = self.config.getlistfloat(
            "simulation", "data_variables_min", fallback=None
        )
        upper_bounds = self.config.getlistfloat(
            "simulation", "data_variables_max", fallback=None
        )
        if lower_bounds is None:
            lower_bounds = [float("nan")] * 4

        if upper_bounds is None:
            upper_bounds = [float("nan")] * 4

        if not len(lower_bounds) == len(upper_bounds) == len(self.data_variables):
            raise ValueError(
                "If bounds are provided, the must be provided for all data "
                "variables. If a bound for a variable is unknown, write 'nan' "
                "in the config file at the position of the variable. "
                "\nE.g.:"
                "\ndata_variables = A B C"
                "\ndata_variables_max = 4 nan 2"
                "\ndata_variables_min = 0 0 nan"
            )
        
        return np.array(lower_bounds), np.array(upper_bounds)

    @property
    def objective(self):
        return self.config.get("inference", "objective_function", fallback="total_average")

    @property
    def n_objectives(self):
        return self.config.getint("inference", "n_objectives", fallback=1)

    @property
    def objective_names(self):
        return self._objective_names

    @property
    def n_threads(self):
        return self.config.getint("multiprocessing", "threads", fallback=4)
    
    @property
    def n_cores(self):
        cpu_avail = mp.cpu_count()
        cpu_set = self.config.getint("multiprocessing", "cores", fallback=1)
        if cpu_set <= 0:
            return cpu_avail + cpu_set
        else: 
            return cpu_set
    
    @n_cores.setter
    def n_cores(self, value):
        self.config.set("multiprocessing", "cores", str(value))

    def create_random_integers(self, n):
        return self.RNG.integers(0, 1e8, n).tolist()
        
    def refill_consumed_seeds(self):
        n_seeds_left = len(self._random_integers)
        if n_seeds_left == self.n_cores:
            n_new_seeds = self._seed_buffer_size - n_seeds_left
            new_seeds = self.create_random_integers(n=n_new_seeds)
            self._random_integers.extend(new_seeds)
            print(f"Appended {n_new_seeds} new seeds to sim.")
        
    def draw_seed(self):
        # return None       
        # the collowing has no multiprocessing stability when the simulation is
        # serialized directly
        self.refill_consumed_seeds()
        seed = self._random_integers.pop(0)
        return seed

    @property
    def seed(self):
        return int(self.config.get("simulation", "seed", fallback=1))

    def set_free_model_parameters(self):
        if self.config.has_section("model-parameters"):
            warnings.warn(
                "config section 'model-parameters' is deprecated, "
                "use 'free-model-parameters' and 'fixed-model-parameters'", 
                DeprecationWarning
            )
            params = parse_config_section(
                self.config["model-parameters"], method="strfloat"
            )
        elif self.config.has_section("free-model-parameters"):
            params = parse_config_section(
                self.config["free-model-parameters"], method="strfloat"
            )
        else:
            warnings.warn("No parameters were specified.")
            params = {}

        
        # create a nested dictionary from model parameters
        parameter_dict = {}
        for par_key, par_value in params.items():
            dp.new(parameter_dict, par_key, par_value, separator=".")

        # create Param instances
        parameters = []
        for param_name, param_dict in parameter_dict.items():
            p = FloatParam(
                value=param_dict.get("value", 1),
                name=param_name,
                min=param_dict.get("min", None),
                max=param_dict.get("max", None),
                step=param_dict.get("step", None),
                prior=param_dict.get("prior", None)
            )
            parameters.append(p)

        return parameters
        
    @property
    def error_model(self):
        em = parse_config_section(self.config["error-model"], method="strfloat")
        return em

    @property
    def evaluator_dim_order(self):
        return self.config.getlist("simulation", "evaluator_dim_order", fallback=None)

    def create_dim_index(self):
        # TODO: If a dimensionality config seciton is implemented this function
        # may become superflous
        sim_dims = self.dimensions
        evaluator_dims = self.evaluator_dim_order
        obs_ordered = self.observations.transpose(*sim_dims)

        var_dim_mapper = {}
        for var in self.data_variables:
            obs_var_dims = obs_ordered[var].dims
            var_dim_mapper.update({
                var: [obs_var_dims.index(e_i) for e_i in evaluator_dims if e_i in obs_var_dims]
            })

        return var_dim_mapper
    
    @property
    def data_structure(self):
        # TODO: If a dimensionality config seciton is implemented this function
        # may become superflous
        obs_ordered = self.observations.transpose(*self.dimensions)

        data_structure = {}
        for var in self.data_variables:
            obs_var_dims = obs_ordered[var].dims
            data_structure.update({
                var: list(obs_var_dims)
            })

        return data_structure

    def reorder_dims(self, Y):
        results = {}
        for var, mapper in self.var_dim_mapper.items():
            results.update({
                var: Y[var][np.array(mapper)]
            })
    
        return results



from functools import partial, lru_cache
from typing import Tuple, Dict, Union, Optional, Callable
import numpyro
import jax
import jax.numpy as jnp
import numpy as np
from numpyro import distributions as dist
from numpyro.infer import Predictive
from numpyro.distributions import Normal, transforms, TransformedDistribution
from numpyro import infer
import xarray as xr
import arviz as az
from matplotlib import pyplot as plt
from sympy import sympify
import sympy2jax

from pymob.simulation import SimulationBase


def LogNormalTrans(loc, scale):
    return TransformedDistribution(
        Normal(0,1), 
        [
            transforms.AffineTransform(loc=jnp.log(loc), scale=scale), 
            exp()
        ]
    )


def generate_transform(expression_str):
    # check for parentheses in expression

    # Parse the expression without knowing the symbol names in advance
    parsed_expression = sympify(expression_str, evaluate=False)
    free_symbols = tuple(parsed_expression.free_symbols)

    # Transform expression to jax expression
    func = sympy2jax.SymbolicModule(parsed_expression, extra_funcs=None, make_array=True)

    return {"transform": func, "args": [str(s) for s in free_symbols]}

exp = transforms.ExpTransform
sigmoid = transforms.SigmoidTransform
C = transforms.ComposeTransform

class NumpyroBackend:
    def __init__(
        self, 
        simulation: SimulationBase
    ):
        """Initializes the NumpyroBackend with a Simulation object.

        Parameters
        ----------
        simulation : SimulationBase
            An initialized simulation.
        """
        self.EPS = 1e-8
        self.distribution_map = {
            "lognorm": (LogNormalTrans, {"scale": "loc", "s": "scale"}),
            "binom": (dist.Binomial, {"n":"total_count", "p":"probs"}),
            "normal": dist.Normal,
            "halfnorm": dist.HalfNormal,
            "poisson": (dist.Poisson, {"mu": "rate"}),
        }
        
        self.simulation = simulation
        self.evaluator = self.parse_deterministic_model()
        self.prior = self.parse_model_priors()
        self.error_model = self.parse_error_model()
        self.inference_model = self.parse_probabilistic_model()

        self.idata: az.InferenceData

    @property
    def plot_function(self) -> Optional[str]:
        return self.simulation.config.get(
            "inference", "plot_function", 
            fallback=None
        )

    @property
    def extra_vars(self):
        extra = self.simulation.config.getlist( 
            "inference", "extra_vars", fallback=[]
        )
        return extra if isinstance(extra, list) else [extra]
    
    @property
    def n_predictions(self):
        return self.simulation.config.getint(
            "inference", "n_predictions", fallback=None
        )
    

    @property
    def chains(self):
        return self.simulation.config.getint(
            "inference.numpyro", "chains", fallback=1
        )
    
    @property
    def draws(self):
        return self.simulation.config.getint(
            "inference.numpyro", "draws", fallback=1000
        )
    
    @property
    def warmup(self):
        return self.simulation.config.getint(
            "inference.numpyro", "warmup", fallback=self.draws
        )

    @property
    def init_strategy(self):
        strategy = self.simulation.config.get(
            "inference.numpyro", "init_strategy", fallback="init_to_median"
        )

        return getattr(infer, strategy)

    def parse_deterministic_model(self) -> Callable:
        """Parses an evaluation function from the Simulation object, which 
        takes a single argument theta and defaults to passing no seed to the
        deterministic evaluator.

        Returns
        -------
        callable
            The evaluation function
        """
        def evaluator(theta, seed=None):
            evaluator = self.simulation.dispatch(theta=theta)
            evaluator(seed)
            return evaluator.Y
        
        return evaluator

    def model(self):
        pass


    def observation_parser(self) -> Tuple[Dict,Dict]:
        """Transform a xarray.Dataset into a dictionary of jnp.Arrays. Creates
        boolean arrays of masks for nan values (missing values are tagged False)

        Returns
        -------
        Tuple[Dict,Dict]
            Dictionaries of observations (data) and masks (missing values)
        """
        obs = self.simulation.observations \
            .transpose(*self.simulation.dimensions)
        data_vars = self.simulation.data_variables + self.extra_vars

        masks = {}
        observations = {}
        for d in data_vars:
            o = jnp.array(obs[d].values)
            m = jnp.logical_not(jnp.isnan(o))
            observations.update({d:o})
            masks.update({d:m})
        
        return observations, masks
    
    @staticmethod
    def parse_parameter(parname: str, prior: str, distribution_map: Dict[str,Tuple]) -> Tuple[str,dist.Distribution,Dict[str,Callable]]:

        distribution, cluttered_arguments = prior.split("(", 1)
        param_strings = cluttered_arguments.split(")", 1)[0].split(",")

        distribution_mapping = distribution_map[distribution]

        if not isinstance(distribution_mapping, tuple):
            distribution = distribution_mapping
            distribution_mapping = (distribution, {})
        
        assert len(distribution_mapping) == 2, (
            "distribution and parameter mapping must be "
            "a tuple of length 2."
        )

        distribution, parameter_mapping = distribution_mapping
        mapped_params = {}
        for parstr in param_strings:
            key, val = parstr.split("=")
            
            mapped_key = parameter_mapping.get(key, key)

            # if is_number(val):
            #     parsed_val = float(val)
            # else:
            parsed_val = generate_transform(expression_str=val)

            # parsed_val = float(val) if is_number(val) else val
            mapped_params.update({mapped_key:parsed_val})

        return parname, distribution, mapped_params

    def parse_model_priors(self):
        priors = {}
        for par in self.simulation.free_model_parameters:
            name, distribution, params = self.parse_parameter(
                parname=par.name, 
                prior=par.prior, 
                distribution_map=self.distribution_map
            )
            # parameterized_dist = distribution(**params)
            priors.update({
                name: {
                    "fn":distribution, 
                    "parameters": params
                }
            })
        return priors

    def parse_error_model(self):
        error_model = {}
        for data_var, error_distribution in self.simulation.error_model.items():
            name, distribution, parameters = self.parse_parameter(
                parname=data_var,
                prior=error_distribution,
                distribution_map=self.distribution_map
            )

                
            error_model.update({
                data_var: {
                    "fn": distribution, 
                    "parameters": parameters,
                    "error_dist": error_distribution
                }
            })
        return error_model

    def parse_probabilistic_model(self):
        EPS = self.EPS
        prior = self.prior.copy()
        error_model = self.error_model.copy()
        extra = {"EPS": EPS}

        def lookup(val, deterministic, prior_samples, observations):
            if val in deterministic:
                return deterministic[val]
            
            elif val in prior_samples:
                return prior_samples[val]
            
            elif val in observations:
                return observations[val]
            
            elif val in extra:
                return extra[val]

            else:
                return val

        def model(solver, obs, masks):
            # construct priors with numpyro.sample and sample during inference
            theta = {}
            for prior_name, prior_kwargs in prior.items():
                
                # apply transforms to priors. This will be handy when I use nested
                # parameters
                prior_distribution_parameters = {}
                for pri_par, pri_val in prior_kwargs["parameters"].items():
                    prior_trans_func = pri_val["transform"]
                    prior_trans_func_kwargs = {k: lookup(k, {}, theta, obs) for k in pri_val["args"]}
                    prior_distribution_parameters.update({
                        pri_par: prior_trans_func(**prior_trans_func_kwargs)
                    })

                theta_i = numpyro.sample(
                    name=prior_name,
                    fn=prior_kwargs["fn"](**prior_distribution_parameters)
                )

                theta.update({prior_name: theta_i})
            
            # calculate deterministic simulation with parameter samples
            sim_results = solver(theta=theta)

            # store data_variables as deterministic model output
            for deterministic_name, deterministic_value in sim_results.items():
                _ = numpyro.deterministic(
                    name=deterministic_name, 
                    value=deterministic_value
                )

            for error_model_name, error_model_kwargs in error_model.items():
                error_distribution = error_model_kwargs["fn"]
                error_distribution_kwargs = {}
                for errdist_par, errdist_val in error_model_kwargs["parameters"].items():
                    errdist_trans_func = errdist_val["transform"]
                    errdist_trans_func_kwargs = {
                        k: lookup(k, sim_results, theta, obs) for k in errdist_val["args"]
                    }
                    error_distribution_kwargs.update({
                        errdist_par: errdist_trans_func(**errdist_trans_func_kwargs)
                    })

                _ = numpyro.sample(
                    name=error_model_name + "_obs",
                    fn=error_distribution(**error_distribution_kwargs).mask(masks[error_model_name]),
                    obs=obs[error_model_name]
                )


            # TODO: How to add EPS
            # DONE: add biomial n --> I think this is already done
            # numpyro.sample("cext_obs", LogNormalTrans(loc=cext + EPS, scale=sigma_cext).mask(masks["cext"]), obs=obs["cext"] + EPS)
            # numpyro.sample("cint_obs", LogNormalTrans(loc=cint + EPS, scale=sigma_cint).mask(masks["cint"]), obs=obs["cint"] + EPS)
            # numpyro.sample("nrf2_obs", LogNormalTrans(loc=nrf2 + EPS, scale=sigma_nrf2).mask(masks["nrf2"]), obs=obs["nrf2"] + EPS)
            # numpyro.sample("lethality_obs", dist.Binomial(probs=leth, total_count=obs["nzfe"]).mask(masks["lethality"]), obs=obs["lethality"])
        return model

    def run(self):
        # set parameters of JAX and numpyro
        # jax.config.update("jax_enable_x64", True)
        numpyro.set_host_device_count(self.chains)

        # generate random keys
        key = jax.random.PRNGKey(1)
        key, *subkeys = jax.random.split(key, 20)
        keys = iter(subkeys)

        # parse observations and masks for missing data
        obs, masks = self.observation_parser()

        # prepare model and print information about shapes
        model = partial(
            self.inference_model, 
            solver=self.evaluator, 
            obs=obs, 
            masks=masks
        )    

        with numpyro.handlers.seed(rng_seed=1):
            trace = numpyro.handlers.trace(model).get_trace()
        print(numpyro.util.format_shapes(trace))
        
        # set up kernel, MCMC        
        kernel = infer.NUTS(
            model, 
            dense_mass=True, 
            step_size=0.01,
            adapt_mass_matrix=True,
            adapt_step_size=True,
            max_tree_depth=10,
            target_accept_prob=0.8,
            init_strategy=self.init_strategy
        )

        mcmc = infer.MCMC(
            sampler=kernel,
            num_warmup=self.warmup,
            num_samples=self.draws,
            num_chains=self.chains,
            progress_bar=True,
        )
    
        # run inference
        mcmc.run(next(keys))
        mcmc.print_summary()

        # create arviz idata
        idata = az.from_numpyro(
            mcmc, 
            dims=self.posterior_data_structure,
            coords=self.posterior_coordinates,
        )
        self.idata = idata

    @property
    def posterior(self):
        return self.idata.posterior


    @property
    def posterior_data_structure(self):
        data_structure = self.simulation.data_structure.copy()
        data_structure_loglik = {f"{dv}_obs": dims for dv, dims in data_structure.items()}
        data_structure.update(data_structure_loglik)
        return data_structure
    
    @property
    def posterior_coordinates(self):
        posterior_coords = self.simulation.coordinates.copy()
        posterior_coords.update({
            "draw": list(range(self.draws)), 
            "chain": list(range(self.chains))
        })
        return posterior_coords
    
    def prior_predictive_checks(self, seed=1):
        key = jax.random.PRNGKey(seed)
        model = partial(self.inference_model, solver=self.evaluator)    
            
        obs, masks = self.observation_parser()

        prior_predictive = Predictive(model, num_samples=100, batch_ndims=2)
        prior_predictions = prior_predictive(key, obs=obs, masks=masks)

        loglik = numpyro.infer.log_likelihood(
            model=model, 
            posterior_samples=prior_predictions, 
            batch_ndims=2, 
            obs=obs, 
            masks=masks
        )

        preds = self.simulation.data_variables
        priors = list(self.simulation.model_parameter_dict.keys())
        posterior_coords = self.posterior_coordinates
        data_structure = self.posterior_data_structure
        
        idata = az.from_dict(
            observed_data=obs,
            prior={k: v for k, v in prior_predictions.items() if k in priors},
            prior_predictive={k: v for k, v in prior_predictions.items() if k in preds},
            log_likelihood=loglik,
            dims=data_structure,
            coords=posterior_coords,
        )

        self.idata = idata
        self.idata.to_netcdf(f"{self.simulation.output_path}/numpyro_prior_predictions.nc")
    
    @staticmethod
    def get_dict(group: xr.Dataset):
        data_dict = group.to_dict()["data_vars"]
        return {k: np.array(val["data"]) for k, val in data_dict.items()}


    @lru_cache
    def posterior_predictions(self, n: Optional[int]=None, seed=1):
        # TODO: It may be necessary that the coordinates should be passed as 
        # constant data. Because if the model is compiled with them once, 
        # according to the design philosophy of JAX, the model will not 
        # be evaluated again. But considering that the jitted functions do take
        # coordinates as an input argument, maybe I'm okay. This should be
        # tested.
        posterior = self.idata.posterior
        stacked_posterior = posterior.stack(sample=("chain", "draw"))
        n_samples = len(stacked_posterior.sample)
    
        if n is not None:
            key = jax.random.PRNGKey(seed)
            selection = jax.random.choice(key=key, a=jnp.array((range(n_samples))), replace=False, shape=(n, ))
            stacked_posterior = stacked_posterior.isel(sample=selection)


        preds = []
        for i in stacked_posterior.sample:
            sample = i.values.tolist()
            chain, draw = sample
            theta = stacked_posterior.sel(sample=sample)
            evaluator = self.simulation.dispatch(theta=self.get_dict(theta))
            evaluator()
            ds = evaluator.results

            ds = ds.assign_coords({"chain": chain, "draw": draw})
            ds = ds.expand_dims(("chain", "draw"))
            preds.append(ds)


        # key = jax.random.PRNGKey(seed)
        # model = partial(self.model, solver=self.evaluator)    
        # predict = numpyro.infer.Predictive(model, posterior_samples=posterior, batch_ndims=2)
        # predict(key, obs=obs, masks=masks)
        

        return xr.combine_by_coords(preds)

    def store_results(self):
        self.idata.to_netcdf(f"{self.simulation.output_path}/numpyro_posterior.nc")

    def load_results(self):
        self.idata = az.from_netcdf(f"{self.simulation.output_path}/numpyro_posterior.nc")

    def plot(self):
        # TODO: combine prior_predictions and posterior predictions
        if hasattr(self.idata, "posterior"):
            axes = az.plot_trace(
                self.idata,
                var_names=self.simulation.model_parameter_names
            )
            fig = plt.gcf()
            fig.savefig(f"{self.simulation.output_path}/trace.png")
            plt.close()
            axes = az.plot_pair(
                self.idata, 
                divergences=True, 
                var_names=self.simulation.model_parameter_names
            )
            fig = plt.gcf()
            fig.savefig(f"{self.simulation.output_path}/pairs_posterior.png")
            plt.close()

        if hasattr(self.idata, "prior_predictive"):
            self.plot_prior_predictive()

        plot_func = getattr(self.simulation, self.plot_function)
        plot_func()

    def plot_prior_predictive(self):
        self.idata

    def plot_predictions(
            self, data_variable: str, x_dim: str, ax=None, subset={}
        ):
        obs = self.simulation.observations.sel(subset)
        
        post_pred = self.posterior_predictions(
            n=self.n_predictions, 
            # seed only controls the parameters samples drawn from posterior
            seed=self.simulation.seed
        ).sel(subset)

        hdi = az.hdi(post_pred, .95)

        if ax is None:
            ax = plt.subplot(111)
        
        y_mean = post_pred[data_variable].mean(dim=("chain", "draw"))
        ax.plot(
            post_pred[x_dim].values, y_mean.values, 
            color="black", lw=.8
        )

        ax.fill_between(
            post_pred[x_dim].values, *hdi[data_variable].values.T, 
            alpha=.5, color="grey"
        )

        ax.plot(
            obs[x_dim].values, obs[data_variable].values, 
            marker="o", ls="", ms=3
        )
        
        ax.set_ylabel(data_variable)
        ax.set_xlabel(x_dim)

        return ax



import json
from configparser import ConfigParser

CONSTANTS = ConfigParser()
CONSTANTS.read("config/constants.cfg")

def read_config(config_file):
    with open(config_file, "r") as f:
        return json.load(f)

def simulation_io_adapter(input_config_file, input_events_file, output_path):
    """
    A simple adapter to adapt the configuration file to be usable with the
    case-study design for the example of experiment based simulation
    """
    
    cfg = read_config(config_file=input_config_file)

    cfg["experiment"]["eventfile"] = input_events_file
    cfg["simulation"]["output"] = output_path

    return cfg


import click
import resource

from pymob.utils import help
from pymob.utils.store_file import prepare_casestudy, import_package

@click.command()
@click.option("-c", "--case_study", type=str, default="test_case_study", 
              help=help.case_study)
@click.option("-s", "--scenario", type=str, default="test_scenario", 
              help=help.scenario)
@click.option("-p", "--package", type=str, default="case_studies", 
              help=help.package)
@click.option("-n", "--n_cores", type=int, default=2, 
              help="The number of cores to be used for multiprocessing")
@click.option("--inference_backend", type=str, default="pymoo")
def main(case_study, scenario, package, inference_backend, n_cores):
    
    config = prepare_casestudy(
        case_study=(case_study, scenario), 
        config_file="settings.cfg", 
        pkg_dir=package
    )

    if n_cores is not None: config.set("multiprocessing", "cores", str(n_cores))
    
    # import package        
    pkg = import_package(package_path=config["case-study"]["package"])
    Simulation = pkg.sim.Simulation
    sim = Simulation(config)

    sim.set_inferer(backend=inference_backend)
    sim.inferer.run()
    sim.inferer.store_results()
    sim.inferer.plot()

    max_ram_mb = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss / 1000
    print("RESOURCE USAGE")
    print("==============")
    print(f"Max RSS: {max_ram_mb} M")


if __name__ == "__main__":
    main()


from typing import Callable, Dict, List
import inspect
import xarray as xr
import numpy as np

def create_dataset_from_numpy(Y, Y_names, coordinates):
    n_vars = Y.shape[-1]
    n_dims = len(Y.shape)
    assert n_vars == len(Y_names), (
        "The number of datasets must be the same as the specified number"
        "of data variables declared in the `settings.cfg` file."
    )

    # transpose Y to put the variable dimension first, then add the
    # remaining dimensions in order
    Y_transposed = Y.transpose((n_dims - 1, *range(n_dims - 1)))

    data_arrays = []
    for y, y_name in zip(Y_transposed, Y_names):
        da = xr.DataArray(y, coords=coordinates, name=y_name)
        data_arrays.append(da)

    dataset = xr.merge(data_arrays)

    return dataset

def create_dataset_from_dict(Y: dict, data_structure, coordinates):
    arrays = {}
    for k, v in Y.items():
        dims = data_structure[k]
        coords = {d: coordinates[d] for d in dims}
        da = xr.DataArray(v, coords=coords, dims=dims)
        arrays.update({k: da})

    return xr.Dataset(arrays)

class Evaluator:
    """The Evaluator is an instance to evaluate a model. It's purpose is primarily
    to create objects that can be spawned and evaluated in parallel and can 
    individually track the results of a simulation or a parameter inference
    process. If needed the evaluations can be tracked and results can later
    be collected.
    
    Seed may not be set as a property, because this should be something passed
    through
    """
    result: xr.Dataset

    def __init__(
            self,
            model: Callable,
            solver: Callable,
            parameters: Dict,
            dimensions: List,
            var_dim_mapper: List,
            data_structure: Dict,
            coordinates: Dict,
            data_variables: List,
            stochastic: bool,
            **kwargs
        ) -> None:
        
        self.model = model
        self.parameters = parameters
        self.dimensions = dimensions
        self.var_dim_mapper = var_dim_mapper
        self.data_structure = data_structure
        self.data_variables = data_variables
        self.coordinates = coordinates
        self.is_stochastic = stochastic
                
        # set additional arguments of evaluator
        _ = [setattr(self, key, val) for key, val in kwargs.items()]

        self._signature = {}

        if isinstance(solver, type):
            self._solver = solver()
        else:
            self._solver = solver

        self.get_call_signature()



    def get_call_signature(self):
        signature = inspect.signature(self._solver)
        model_args = list(signature.parameters.keys())

        for a in model_args:
            if a not in self.allowed_model_signature_arguments:
                raise ValueError(
                    f"{a} in model signature are not attributes of the Evaluator. "
                    f"Use one of {self.allowed_model_signature_arguments}, "
                    f"or set as evaluator_kwargs in the call to "
                    "'SimulationBase.dispatch'" 
                )
            
            # add argument to signature for call to model
            if a != "seed":
                self._signature.update({a: getattr(self, a)})
        
    
    @property
    def allowed_model_signature_arguments(self):
        return [a for a in self.__dict__.keys() if a[0] != "_"] + ["seed"]

    def __call__(self, seed=None):
        if seed is not None:
            self._signature.update({"seed": seed})

        self.Y = self._solver(**self._signature)

    @property
    def dimensionality(self):
        return {key: len(values) for key, values in self.coordinates.items()}

    @property
    def results(self):
        if isinstance(self.Y, dict):
            return create_dataset_from_dict(
                Y=self.Y, 
                coordinates=self.coordinates,
                data_structure=self.data_structure,
            )
        elif isinstance(self.Y, np.ndarray):
            return create_dataset_from_numpy(
                Y=self.Y,
                Y_names=self.data_variables,
                coordinates=self.coordinates,
            )
        else:
            raise NotImplementedError(
                "Results returned by the solver must be of type Dict or np.ndarray."
            )
    